/*
 * Copyright (c) 1998 PFU Limited.
 *	author Osamu Satoh, Kazuhisa Shirakami
 */
package teikade.util;

/**
 * The class represents compilation result.
 */
public class CompileResult {
    /**
     * Symbol indicates a message was detected.
     * @see type()
     * @see CompileResultListener#handleResult(CompileResult)
     */
    public static final int MESSAGE = 0;

    /**
     * Symbol indicates a warning was detected.
     * @see type()
     * @see CompileResultListener#handleResult(CompileResult)
     */
    public static final int WARNING = 1;

    /**
     * Symbol indicates an error was detected.
     * @see type()
     * @see CompileResultListener#handleResult(CompileResult)
     */
    public static final int ERROR = 2;

    /**
     * Symbol indicates something unrecoverable was happened.
     * @see type()
     * @see CompileResultListener#handleResult(CompileResult)
     */
    public static final int FATAL = 3;

    private int type;
    private String message;
    private String fileName;
    private int line;
    private int offset;
    private String description;

    public CompileResult(String message) {
	type = MESSAGE;
	this.message = message;
    }

    public CompileResult(int type, String fileName, int line, int offset,
			 String description, String message) {
	this.type = type;
	this.fileName = fileName;
	this.line = line;
	this.offset = offset;
	this.description = description;
	this.message = message;
    }

    /**
     * Creates a message.
     * @param message message
     */
    public static CompileResult createMessage(String message) {
	return new CompileResult(message);
    }

    /**
     * Creates a warning.
     * @param fname file name on which this error was detected
     * @param line line number where this warning happened
     *        0 means no appropriate line for this warning
     * @param offset offset in the description
     *        0 means no offset
     * @param desc short description of this warning
     * @param msg full description of this warning
     */
    public static CompileResult createWarning(String fname, int line, int col, String desc, String msg) {
	return new CompileResult(WARNING, fname, line, col, desc, msg);
    }

    /**
     * Creates an error.
     * @param fname file name on which this error was detected
     * @param line line number where this error happened
     *        -1 means no appropriate line for this error
     * @param offset offset in description
     *        0 means no offset
     * @param desc short description of this error
     * @param msg full description of this error
     */
    public static CompileResult createError(String fname, int line, int col, String desc, String msg) {
	return new CompileResult(ERROR, fname, line, col, desc, msg);
    }

    /**
     * Creates a fatal error.
     * @param fname file name on which this error was detected
     * @param line line number where error happened
     *        -1 means no appropriate line for this error
     * @param offset offset in description
     *        0 means no offset
     * @param desc short description of this error
     * @param msg full description of this error
     */
    public static CompileResult createFatal(String fname, int line, int col, String desc, String msg) {
	return new CompileResult(FATAL, fname, line, col, desc, msg);
    }

    /**
     * Returns type of this result.
     * @see MESSAGE
     * @see WARNING
     * @see ERROR
     * @see FATAL
     */
    public final int type() {
	return type;
    }

    /**
     * Returns the file name where error was detected.
     */
    public final String fileName() {
	return fileName;
    }

    /**
     * Returns the line number where error was detected.
     * It is offset from the top of sources.
     */
    public final int line() {
	return line;
    }

    /**
     * Returns the offse in the descripton.
     * It is offset from the top of line.
     */
    public final int offset() {
	return offset;
    }

    /**
     * Returns brief description of the error.
     */
    public final String description() {
	return description;
    }

    /**
     * Returns full detailed error description.
     */
    public final String message() {
	return message;
    }

    public String toString() {
	return getClass().getName()
	     + "[fileName=" + fileName()
	     + ",line=" + line()
	     + ",offset=" + offset()
	     + ",description=" + description()
	     + ",message=" + message()
	     + ",type=" + type()
	     + "]";
    }
}
