/*
 * Copyright (c) 1995 PFU Limited.
 *      author Kazuhisa Shirakami
 */
package teikade.util;

import teikade.sys.TeikadeProperties;
import teikade.sys.CodePath;
import teikade.sys.SystemManager;
import teikade.sys.ProjectManager;
import teikade.sys.PackageManager;
import java.util.Vector;
import java.util.Enumeration;

public abstract class Grep implements Command, Runnable {
    public static final String USE_INTERNAL = "grep.internal";
    public static final String INTERNAL_CLASS_NAME = "grep.internal.class";
    public static boolean useInternalGrep =
        (SystemManager.onWin32 || SystemManager.onOS2) ?
	    !"false".equals(TeikadeProperties.getString(USE_INTERNAL)) :
	    "true".equals(TeikadeProperties.getString(USE_INTERNAL));
    public static boolean canExecute = true;
    static {
	TeikadeProperties.setDefault(INTERNAL_CLASS_NAME, "teikade.util.GrepInternalExpression");
    }

    private Thread thread;
    private CodePath codePath;
    private String pattern;
    private GrepListener listener;

    public static boolean canExecute() {
	return canExecute;
    }

    public static Grep getInstance(CodePath cp,
				   String pattern,
				   GrepListener listener) {
	Grep grep = getInstance();
	grep.setCodePath(cp);
	grep.setPattern(pattern);
	grep.setGrepListener(listener);
	return grep;
    }

    public static Grep getInstance() {
	if (useInternalGrep) {
	    String cls = TeikadeProperties.getString(INTERNAL_CLASS_NAME);
	    if (cls != null && cls.length() > 0) {
		try {
		    return (Grep)Class.forName(cls).newInstance();
		} catch (ClassCastException e) {
		    SystemManager.information("grep: the class "
			+ cls + " is not subclass of teikade.util.Grep"
		        + "\nin the properties for grep."
			+ "\nUse default class for grep");
		} catch (Exception e) {
		    SystemManager.information("grep: fail to instanciate the class " + cls
			+ e.getClass().getName()
			+ "\nin the properties for grep."
			+ "\nUse default class for grep");
		}
	    }
	    return new GrepInternalExpression();
	} else {
	    return new GrepExternal();
	}
    }

    public void setCodePath(CodePath codePath) {
	this.codePath = codePath;
    }

    public CodePath getCodePath() {
	return codePath;
    }

    public void setGrepListener(GrepListener listener) {
	this.listener = listener;
    }

    public void setPattern(String pattern) {
	this.pattern = pattern;
    }

    public String getPattern() {
	return pattern;
    }

    public void execute() {
	abort();
	thread = new Thread(this,this.getClass().getName());
	thread.start();
    }

    public void abort() {
	if (thread != null) {
	    thread.interrupt();
	    thread = null;
	}
    }

    public void executing() {
	listener.executing(this);
    }

    public void finished(int status) {
	thread = null;
	listener.finished(this, status);
    }

    public void aborted(Exception e) {
	thread = null;
	listener.aborted(this, e);
    }

    public void addResult(CodeLocation result) {
	listener.result(this, result);
    }

    public void addMessage(String message) {
	listener.message(this, message);
    }

    public String getPatternLabel() {
	return "egrep pattern:";
    }

    protected Vector getSources(CodePath cp) {
	ProjectManager prj;
	PackageManager pkg;
	Vector files = new Vector();
	if (cp.classManager() != null) {
	    files.addElement(cp.sourceFileManager());
	} else if ((pkg = cp.packageManager()) != null) {
	    addFiles(files, pkg);
	} else if ((prj = cp.projectManager()) != null) {
	    Enumeration pkgs = prj.packages();
	    while (pkgs.hasMoreElements()) {
		 addFiles(files, (PackageManager)pkgs.nextElement());
	    }
	} else {
	    Enumeration projs = SystemManager.allProjects();
	    while (projs.hasMoreElements()) {
		Enumeration pkgs = prj.packages();
		while (pkgs.hasMoreElements()) {
		    addFiles(files, (PackageManager)pkgs.nextElement());
		}
	    }
	}
	return files;
    }

    protected static void addFiles(Vector files, PackageManager pkg) {
       Enumeration srcs = pkg.sources();
       while (srcs.hasMoreElements()) {
          files.addElement(srcs.nextElement());
       }
    }
}
