/*
 * Copyright (c) 1995 PFU Limited.
 *	author Osamu Satoh, Kazuhisa Shirakami
 */
package teikade.util;

import teikade.sys.*;
import java.util.*;
import java.io.*;

public class GrepExternal extends Grep implements CommandExternalListener {
    private static final String egrepChars = ".*^$[]+?|";
    CommandExternal command;
    String separator;
    char separatorChar;

    public void run() {
	try {
	    command = new CommandExternal(makeCommand(), this);
	    command.execute();
	} catch (TeikadeException e) {
	    aborted(e);
	}
    }

    protected String[] makeCommand() throws TeikadeException {
	String command   = TeikadeProperties.getString("command.grep");
	separator = TeikadeProperties.getString("command.grep.separator");
	String quote     = TeikadeProperties.getString("command.grep.quote");
	if (SystemManager.onWin32 || SystemManager.onOS2) {
	    if (command   == null) command = "egrep.exe -n -e %pattern% nul";
	    if (separator == null) separator = "/";
	    if (quote     == null) quote = "\"";
	} else {
	    if (command   == null) command = "egrep -n -e %pattern% /dev/null";
	    if (separator == null) separator = "/";
	    if (quote     == null) quote = "'";
	}
	separatorChar = separator.charAt(0);
	if (separatorChar == File.separatorChar) separator = null;
	boolean hasPattern = false;
	Vector commands = new Vector();
	StringTokenizer tokens = new StringTokenizer(command);
	if (!tokens.hasMoreTokens()) {
	    throw new TeikadeException("no command in command.grep property");
	}
	commands.addElement(tokens.nextToken());
	while (tokens.hasMoreTokens()) {
	    String token = tokens.nextToken();
	    if ("%pattern%".equals(token)) {
		commands.addElement(quote + getPattern() + quote);
		hasPattern = true;
	    } else {
		commands.addElement(token);
	    }
	}
	if (!hasPattern) {
	    throw new TeikadeException("no \"%pattern%\" parameter in command.grep property");
	}
	int firstFileName = commands.size();
	addFileNames(commands);
	String[] commandArray = new String[commands.size()];
	commands.copyInto(commandArray);
	if (separator != null) {
	    for (int i = firstFileName; i < commandArray.length; i++) {
		commandArray[i] = commandArray[i].replace(File.separatorChar, separatorChar);
	    }
	}
	return commandArray;
    }

    protected void addFileNames(Vector vector) {
        ProjectManager prj;
        PackageManager pkg;
	CodePath cp = getCodePath();
        if (cp.classManager() != null) {
            String fname = cp.sourceFileManager().absolutePath();
            vector.addElement(fname);
        } else if ((pkg = cp.packageManager()) != null) {
	    addFileNames(vector, pkg);
        } else if ((prj = cp.projectManager()) != null) {
	    addFileNames(vector, prj);
        } else {
            Enumeration projs = SystemManager.allProjects();
            while (projs.hasMoreElements()) {
		addFileNames(vector, (ProjectManager)projs.nextElement());
	    }
        }
    }

    protected static void addFileNames(Vector vector, PackageManager pkg) {
	if (SystemManager.onWin32 || SystemManager.onOS2) {
	    vector.addElement(pkg.absolutePath() + File.separator + "*.java");
	} else {
	    Enumeration srcs = pkg.sources();
	    while (srcs.hasMoreElements()) {
		vector.addElement(((SourceFileManager)srcs.nextElement()).absolutePath());
	    }
	}
    }

    protected static void addFileNames(Vector files, ProjectManager prj) {
	Enumeration pkgs = prj.packages();
	while (pkgs.hasMoreElements()) {
	    addFiles(files, (PackageManager)pkgs.nextElement());
	}
    }

    public void executing(Command grep) {
	executing();
    }

    public void finished(Command grep, int status) {
	finished(status);
    }

    public void aborted(Command grep, Exception e) {
	aborted(e);
    }

    public void output(Command command, String line) {
	int first  = line.indexOf(':');
	if ((SystemManager.onWin32 || SystemManager.onOS2) &&
	    first == 1 && (line.charAt(2) == separatorChar ||
			   line.charAt(2) == File.separatorChar)) {
	    char c = line.charAt(0);
	    if ('A' <= c && c <= 'Z' || 'a' <= c && c <= 'z') {
		first = line.indexOf(':', 3);
	    }
	}
	int second = line.indexOf(':', first + 1);
	if (first < 0 && second <= first) {
	    SystemManager.information(line);
	    return;
	}

	String fileName = line.substring(0, first);
	if (separator != null) {
	    fileName = fileName.replace(separatorChar,
					File.separatorChar);
	}
	int lineNumber = 0;
	try {
	    lineNumber = parseNumber(line, first + 1, second);
	} catch (NumberFormatException e) {
	    addMessage(line + "\n");
	    return;
	}

	int offset, length;
	String ss = trimRegexp(getPattern());
	if (ss.charAt(ss.length() - 1) == '$') {
	    offset = line.lastIndexOf(ss);
	} else {
	    offset = line.indexOf(ss, second + 1);
	}
	if (offset <= second) {
	    offset = 0;
	    length = -1;
	} else {
	    offset -= second + 1;
	    length = ss.length();
	}

	int separator = line.lastIndexOf(separatorChar, first);
	String listString =
	       (separator <= 0 ? line : line.substring(separator + 1));

	try {
	    CodeLocation item = new CodeLocation(fileName, lineNumber,
					listString, offset, length);
	    addResult(item);
	} catch (CodeLocationException e) {
	    addMessage("fail to search location " +
		       fileName + ":" + lineNumber + ", sorry.\n");

	}
    }

    public void error(Command command, String line) {
	output(command, line);
    }

    protected String trimRegexp(String pattern) {
	int start, end;
	if ((start = skipEgrepChars(pattern, 0)) >= 0) {
	    if (pattern.charAt(start) == '\\') start++;
	    if ((end = indexOfEgrepChar(pattern, start + 1)) < 0)
		end = pattern.length();
	    return pattern.substring(start, end);
	}
	return pattern;
    }

    private static int parseNumber(String str, int pos, int end) {
	int number = 0;
	while (pos < end) {
	    char c = str.charAt(pos);
	    if (Character.isDigit(c)) {
		number = number * 10 + (c - '0');
	    }
	    pos++;
	}
	return number;
    }

    private static int isEgrepChar(String str, int start) {
	int len = str.length();
	if (start >= len) return -1;
	char c = str.charAt(start);
	if (c == '\\') {
	    int end = start + 1;
	    if (end >= len) return 1;
	    if (str.charAt(end) != '(') return 0;
	    end++;
	    for (;;) {
		end = str.indexOf('\\', end) + 2;
		if (end < 2 || end >= len) return len - start;
		if (str.charAt(end - 1) == ')') return end - start;
	    }
	}
	if (c == '[') {
	    int end = start + 2;
	    if (end >= len) return len - start;
	    for (;;) {
		end = str.indexOf(']', end) + 1;
		if (end <= 0 || end >= len) return len - start;
		if (str.charAt(end - 2) != '\\') return end - start;
	    }
	}
	if (egrepChars.indexOf(c) >= 0) return 1;
	return 0;
    }

    static private int indexOfEgrepChar(String str, int pos) {
	int max = str.length();
	while (pos < max) {
	    int length = isEgrepChar(str, pos);
	    if (length > 0) return pos;
	    if (length < 0) return -1;
	    pos++;
	}
	return -1;
    }

    static private int skipEgrepChars(String str, int pos) {
	int max = str.length();
	while (pos < max) {
	    int length = isEgrepChar(str, pos);
	    if (length < 0) return -1;
	    if (length == 0) return pos;
	    pos += length;
	}
	return -1;
    }
}
