/*
 * Copyright (c) 1998 SCK Files, PFU Limited.
 * author Maurizio Scaglione.
 *
 */
package teikade.util;

import teikade.sys.*;
import java.util.*;
import java.io.FileInputStream;
import java.io.IOException;

public class GrepInternalExpression extends Grep {
    private Vector files;
    private static final String egrepChars = ".*^$[]+?|";
    protected String trimmedPattern;
    protected int patternLength;
    protected GrepPattern head;

    public synchronized void run() {
	executing();
	trimmedPattern = trimRegexp(getPattern());
	patternLength = trimmedPattern.length();
	head = new GrepPattern(CodeConverter.toBytes(getPattern()));
        Enumeration enum = files.elements();
        while (enum.hasMoreElements()) {
            SourceFileManager src = (SourceFileManager)enum.nextElement();
            try {
		FileInputStream fis = new FileInputStream(src.absolutePath());
		byte[] text = new byte[fis.available()];
		fis.read(text);
		fis.close();
		grep(text, src);
            } catch (Exception e) {
		System.err.println("internal grep caught Exception:\n" + e.getMessage());
		aborted(e);
            }
	}
	finished(0);
    }

    protected void grep(byte[] text, SourceFileManager src) {
	
	int bol = 0, eol = 0, line = 1;
	for (int i = 0; i < text.length; i++) {
	    switch (text[i]) {
	    case '\r':		       // CR
		eol = i;
		if (i + 1 < text.length && text[i + 1] == '\n') {
		    ++i;               // CR+LF
		}
		break;
	    case '\n':                 // LF
		eol = i;
		break;
	    default:
		continue;
	    }
	    if (eol > bol) { // check line
		int index = -1;
		if ((index = head.grep(text, bol, eol)) != -1) {
		    try {
			String lineString = CodeConverter.toString(text, bol, eol - bol);
			int offset = lineString.indexOf(trimmedPattern, index - bol);
			String detail = src.path() + ":" + line + ":" + lineString;
			CodeLocation result = new CodeLocation(
			     src.absolutePath(), line, detail,
			     offset, patternLength);
			 addResult(result);
		    } catch (CodeLocationException e) {
			 addMessage("fail toIndex search location "
			 + src.absolutePath() + ":" + line + ", sorry.\n");
		    }
		}
	    }
	    ++line;
	    bol = i + 1;
	}
    }

    protected String trimRegexp(String pattern) {
	int start, end;
	if ((start = skipEgrepChars(pattern, 0)) >= 0) {
	    if (pattern.charAt(start) == '\\') start++;
	    if ((end = indexOfEgrepChar(pattern, start + 1)) < 0)
		end = pattern.length();
	    return pattern.substring(start, end);
	}
	return pattern;
    }

    private static int isEgrepChar(String str, int start) {
	int len = str.length();
	if (start >= len) return -1;
	char c = str.charAt(start);
	if (c == '\\') {
	    int end = start + 1;
	    if (end >= len) return 1;
	    if (str.charAt(end) != '(') return 2;
	    end++;
	    for (;;) {
		end = str.indexOf('\\', end) + 2;
		if (end < 2 || end >= len) return len - start;
		if (str.charAt(end - 1) == ')') return end - start;
	    }
	}
	if (c == '[') {
	    int end = start + 2;
	    if (end >= len) return len - start;
	    for (;;) {
		end = str.indexOf(']', end) + 1;
		if (end <= 0 || end >= len) return len - start;
		if (str.charAt(end - 2) != '\\') return end - start;
	    }
	}
	if (egrepChars.indexOf(c) >= 0) return 1;
	return 0;
    }

    private static int indexOfEgrepChar(String str, int pos) {
	int max = str.length();
	while (pos < max) {
	    int length = isEgrepChar(str, pos);
	    if (length > 0) return pos;
	    if (length < 0) return -1;
	    pos++;
	}
	return -1;
    }

    private static int skipEgrepChars(String str, int pos) {
	int max = str.length();
	while (pos < max) {
	    int length = isEgrepChar(str, pos);
	    if (length < 0) return -1;
	    if (length == 0) return pos;
	    pos += length;
	}
	return -1;
    }

    public void setCodePath(CodePath cp) {
	super.setCodePath(cp);
	files = getSources(cp);
    }
}

final class GrepPattern {
    private byte[] pattern;
    int length;
    private int engine = 0;
    private byte[][] expr;
    
    static final int ON_START = 0x0001;
    static final int ON_END   = 0x0002;

    GrepPattern(byte[] ptrn) {
	pattern = new byte[ptrn.length];
	System.arraycopy(ptrn, 0, pattern, 0, ptrn.length);
	if (pattern[0] == '^') {
	    engine = ON_START;
	    pattern = removeElementAt(pattern, 0);
	} else if(pattern[pattern.length-1] == '$') {
	    engine = ON_END;
	    pattern = removeElementAt(pattern, pattern.length - 1);
	}
	/*
	while (pattern[pattern.length-1] == '*') {
	    if (pattern.length > 1 && pattern[pattern.length-2] != '\\') {
		pattern = removeElements(pattern, pattern.length - 2, 2);
	    }
	} */
	pattern = parse(pattern);
	length = pattern.length;
    }

    byte[] parse(byte[] pattern) {
	int index = 0;
	expr = new byte[pattern.length][0];
	byte[] newPattern = new byte[pattern.length];
	int count = 0;
	for (int i = 0; i < pattern.length; i++) {
	    switch (pattern[i]) {
	    case '\\': // consider special chars as normal
		pattern = removeElementAt(pattern, i);
		break;
	    case '.' : // any byte
		pattern[i] = (byte)-(++index);
		expr[index - 1] = new byte[1];
		expr[index - 1][0] = (byte)'.';
		break;
	    case '+' : // 1 or more repetitions
		pattern[i] = (byte)-(++index);
		expr[index - 1] = new byte[2];
		expr[index - 1][0] = (byte)'*';
		expr[index - 1][1] = pattern[i - 1];
		break;
	    case '?' : // 0 or 1 repetitions
		pattern[i] = (byte)-(++index);
		expr[index - 1] = new byte[2];
		expr[index - 1][0] = (byte)'?';
		--i;
		expr[index - 1][1] = pattern[i];
		pattern = removeElementAt(pattern, i);
		break;
	    case '*' : // 0 or more repetitions
		pattern[i] = (byte)-(++index);
		expr[index - 1] = new byte[2];
		expr[index - 1][0] = (byte)'*';
		--i;
		expr[index - 1][1] = pattern[i];
		pattern = removeElementAt(pattern, i);
		break;
	    case '[' : // multiple chars
		pattern[i] = (byte)-(++index);
		int j = i;
		while (j < pattern.length && pattern[j] != ']') ++j; 
		int size = j - i;
		expr[index - 1] = new byte[size];
		expr[index - 1][0] = (byte)'[';
		System.arraycopy(pattern, i + 1, expr[index - 1], 1, size - 1);
		pattern = removeElements(pattern, i + 1, size);
		break;
	    }
	    if (pattern[i] < 0) {
		System.out.println("expr " + (char)expr[index - 1][0] + " id " +
	                       pattern[i] + CodeConverter.toString(expr[index - 1]));
	    }
	}
	return pattern;
    }

    int grep(byte[] text, int fromIndex, int toIndex) {
	switch (engine) {
	case 0: return indexOf(text, fromIndex, toIndex, this);
	case ON_START: return startsWith(text, fromIndex, toIndex, this);
	case ON_END: return endsWith(text, fromIndex, toIndex, this);
	default: return indexOf(text, fromIndex, toIndex, this);
	}
    }

    int eval(int state, byte t) {
	if (pattern[state] < 0) {
	    return expr(pattern[state], t);
	} else if (t == pattern[state]) return 0;
	else return -1;
    }

    int expr(byte b, byte t) {
	try {
	    int i = -(int)b - 1;
	    byte bb;
	    switch (expr[i][0]) {
	    case '.':
		return 0;
	    case '?':
		bb = expr[i][1];
		if (bb < 0) {
		    return expr(bb, t);
		} else if (t == bb) {
		    return 0;
		}
		return 4;
	    case '*':
		bb = expr[i][1];
		if (bb < 0) {
		    if (expr(bb, t) == 0) return 3;
		} else if (t == bb) {
		    return 1;
		}
		return 2;
	    case '[':
		boolean not = expr[i][1] == '^';
		for (int j = (not) ? 2 : 1; j < expr[i].length; j++) {
		    bb = expr[i][j];
		    if (bb != '\\') {
			if (bb == '-') {
			    if (t > expr[i][j-1] && t < expr[i][j+1]) {
				return (not) ? -1 : 0;
			    }
			} else {
			    if (t == bb) return (not) ? -1 : 0;
			}
		    } else if (++j < expr[i].length) {
			if (t == expr[i][j]) return (not) ? -1 : 0;
		    }
		} 
		return (not) ? 0 : -1;
	    default:
		return -1;
	    }
	} catch (ArrayIndexOutOfBoundsException e) {
	    return -1;
	}
    }

    static final int indexOf(byte[] text, int fromIndex, int toIndex, GrepPattern pattern) {
	if (pattern.length == 0) return 0;
	int any = 0;
	int state = 0, markState = 0, markIndex = 0;
	for (int i = fromIndex; i <= toIndex; i++) {
	    switch(pattern.eval(state, text[i])) {
	    case 3: // eval same but don't reset
		markState = state;
		markIndex = i;
		++state;
		--i;
		break;
	    case 2: // eval same then reset
		if (state == 0) markIndex = i;
		++state;
		--i;
		break;
	    case 1: // go on
		continue;
	    case 0: // ok
		if (state == 0) markIndex = i;
		++state;
		break;
	    case -1: // reset
		if (state > 0) i = markIndex;
		state = markState;
		markState = 0;
		continue;
	    default: // reset
		if (state > 0) i = markIndex;
		state = markState;
		markState = 0;
		continue;
	    }
	    if (state >= pattern.length) return markIndex;
	}
	return -1;
    }

    static final int startsWith(byte[] text, int fromIndex, int toIndex, GrepPattern pattern) {
	if (pattern.length == 0) return 0;
	int any = 0;
	int state = 0, markState = 0, markIndex = 0;
	for (int i = fromIndex; i <= toIndex; i++) {
	    switch(pattern.eval(state, text[i])) {
	    case 3: // eval same but don't reset
		markState = state;
		markIndex = i;
		++state;
		--i;
		continue;
	    case 2: // eval same then reset
		++state;
		--i;
		continue;
	    case 1: // go on
		continue;
	    case 0: // ok
		++state;
		break;
	    case -1: // reset
		if (markIndex > 0) {
		    i = markIndex;
		    state = markState;
		    markState = 0;
		    markIndex = 0;
		    continue;
		} else {
		    return -1;
		}
	    default: // reset
		if (markIndex > 0) {
		    i = markIndex;
		    state = markState;
		    markState = 0;
		    markIndex = 0;
		    continue;
		} else {
		    return -1;
		}
	    }
	    if (state >= pattern.length) return 0;
	}
	return -1;
    }

    static final int endsWith(byte[] text, int fromIndex, int toIndex, GrepPattern pattern) {
	if (pattern.length == 0) return 0;
	int any = 0, m = pattern.length-1;
	int state = m, markState = m, markIndex = toIndex;
	for (int i = toIndex; i >= fromIndex; i--) {
	    switch(pattern.eval(state, text[i])) {
	    case 3: // eval same but don't reset
		markState = state;
		markIndex = i;
		--state;
		++i;
		continue;
	    case 2: // eval same then reset
		--state;
		++i;
		continue;
	    case 1: // go on
		continue;
	    case 0: // ok
		--state;
		break;
	    case -1: // reset
		if (markIndex < toIndex) {
		    i = markIndex;
		    state = markState;
		    markState = m;
		    markIndex = toIndex;
		    continue;
		} else {
		    return -1;
		}
	    default: // reset
		if (markIndex < toIndex) {
		    i = markIndex;
		    state = markState;
		    markState = m;
		    markIndex = toIndex;
		    continue;
		} else {
		    return -1;
		}
	    }
	    if (state < 0) return i;
	}
	return -1;
    }

    static final byte[] removeElementAt(byte[] bytes, int index) {
	return removeElements(bytes, index, 1);
    }

    static final byte[] removeElements(byte[] bytes, int index, int length) {
	byte[] newBytes = new byte[bytes.length - length];
	if (index == 0) {
	    System.arraycopy(bytes, length, newBytes, 0, newBytes.length);
	} else if (index == newBytes.length) {
	    System.arraycopy(bytes, 0, newBytes, 0, newBytes.length);
	} else {
	    System.arraycopy(bytes, 0, newBytes, 0, index);
	    System.arraycopy(bytes, index + length, newBytes, index, newBytes.length - index);
	}
	return newBytes;
    }
}
