package teikade.util;

import java.io.*;

/***************************************************************
ThreadUtil provides a simple mean of tracing and centralizing
usage of deprecated and dangerous Thread calls: suspend/resume
and stop.
The suspend/resume pair has been implemented based on the Sun's
recommended strategy in the paper "Why are Thread.stop, Thread.suspend
and Thread.resume Deprecated?" on their web site java.sun.com.
The stop method has not been reimplemented (yet?), still uses the
deprecated Thread.stop method.

@author Achille M.V. Petrilli
@version 1.0
*****************************************************************/

public final class ThreadUtil {
	// if you suspect this code not to work properly, just set the
	// USE_DEPRECATED to true and recompile. The deprecated methods
	// will be used then.
	private static final boolean USE_DEPRECATED = false;
	private static final PrintStream trace = System.err;
	
	// only static methods, class is not instantiatable
	private ThreadUtil() {
	}
	
	/**************************************************************
	Suspend(Thread t) uses the given Thread object to synchronize
	and wait for a notification, effectively suspending the calling
	Thread, not the given thread!
	On the other side, teikade's usage of Thread.suspend() seems to
	always follow this pattern (experimentally tested)).
	**************************************************************/
	public static void Suspend(Thread t) {
		if (USE_DEPRECATED) {
			t.suspend();
			return;
		}
		// trace only "problem" calls to Suspend (i.e. won't work if trying to suspend any
		//	   other thread than the current one!!!)
		if (t != Thread.currentThread()) trace("suspend",t);
		while (true) {
			try {
				synchronized(t) {
					t.wait();
				}
				break;
			} catch (InterruptedException e) {}
		}
	}

	/**************************************************************
	Resume(Thread t) uses the given Thread object to synchronize
	and send a notification to threads waiting on the given Thread
	object, which means that any thread which may be waiting on that
	Thread object will wake up, not just the Thread itself!!!
	If the Thread is not suspended, no harm is done.
	**************************************************************/
	public static void Resume(Thread t) {
		if (USE_DEPRECATED) {
			t.resume();
			return;
		}
		// trace only "problem" calls to Resume (i.e. can't resume myself)
		if (t == Thread.currentThread()) trace("resume",t);
		synchronized(t) {
			t.notifyAll();
		}
	}
	public static void Stop(Thread t) {
		if (USE_DEPRECATED) {
			t.stop();
			return;
		}
		// trace all calls to Stop
		//trace("stop",t);
		// trace only "problem" calls to Stop (i.e. can't stop myself)
		//if (t != Thread.currentThread()) trace("stop",t);
		t.stop();
	}
	private static void trace(String op, Thread t) {
		trace.println("<<<<< "+Thread.currentThread()+" "+op+"-ing "+t+" begin.");
		Throwable e = new Throwable(Thread.currentThread()+" stack trace:");
		e.fillInStackTrace().printStackTrace(trace);
		trace.println(">>>>> "+Thread.currentThread()+" "+op+"-ing "+t+" done.");
		trace.flush();
	}
}
